import socket
import threading
import logging
import datetime
import os


from Server.utils import prepare_model,receive,save_model,load_input
logging.basicConfig(level=logging.NOTSET)


class DatachannelServer:
    def __init__(self, ip, port,clientsDB,modelsDB,server_model_path,model_type,golden_data_input,golden_data_output,score_fn,gap_time=20, listener_num = 100):
        self.ip = ip
        self.port = port
        self.server_model_path = server_model_path
        self.model_type = model_type
        self.listener_num = listener_num
        self.gap_time = gap_time
        self.clientsDB = clientsDB
        self.modelsDB = modelsDB
        self.server = None
        self.golden_data_input,self.golden_data_output = load_input(golden_data_input,golden_data_output)
        self.score_fn = score_fn
        self.received_values = {}
        logging.debug('Datachannel initialized')
    
        
    def start(self, server_controller):
        if self.server is not None and self.server.fileno() != -1:
            logging.info("Datachannel server is already running")
            return
        self.server_controller = server_controller
        self.last_time = None
        self.barrier = threading.Barrier(2)
        self.received_values = {}
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind((self.ip, self.port))
        self.server.listen(self.listener_num)
        logging.debug('Datachannel server started on {}:{}'.format(self.ip, self.port))
        self.run()

    def run(self):
        counter_thread = threading.Thread(target=self.count_average)
        counter_thread.start()
        logging.debug('counter_thread started')
        time_controller_thread = threading.Thread(target=self.check_time)
        time_controller_thread.start()
        logging.debug('time_controller_thread started')
        main_thread = threading.Thread(target=self.runner_thread)
        main_thread.start()
        logging.debug('main_thread started')

    
    def calculate_average(self):
        if self.model_type == "pytorch":
            import torch
            keys = list(self.received_values.keys())
            print("SDASDASDASDASD",self.received_values)
            if len(keys) == 0:
                return None
            init_weights = self.received_values[keys[0]]
            averaging_weight = self.score_fn(init_weights(self.golden_data_input),self.golden_data_output)
            sum_weighted_average = averaging_weight
            with torch.no_grad():
                for i in init_weights.parameters():
                    i*=averaging_weight
            print("Before")
            for key in keys:
                for i in self.received_values[key].parameters():
                    print(i)
                print("Another Model")
            for mdl in range(1,len(keys)):
                averaging_weight = self.score_fn(self.received_values[keys[mdl]](self.golden_data_input),self.golden_data_output)
                sum_weighted_average += averaging_weight
                with torch.no_grad():
                    for (i,j) in zip(init_weights.parameters(),self.received_values[keys[mdl]].parameters()):
                        i+=averaging_weight*j 
            
            with torch.no_grad():
                for i in init_weights.parameters():
                    i/=sum_weighted_average
            
            print("After")
            for i in init_weights.parameters():
                print(i)
            
            return init_weights
        elif self.model_type == "tensorflow":
            models = list(self.received_values.values())
            if len(models) == 0:
                print("AAABa")
                return None
            import tensorflow as tf
            print("Before")
            for model in models:
                for layer in model.layers:
                    print(layer.get_weights())
                    print("\n")
            model = models[0]
            averaging_weight = self.score_fn(model(self.golden_data_input),self.golden_data_output)
            sum_weighted_average = averaging_weight
            for layer in model.layers:
                curr_weights = layer.get_weights()
                for i in range(len(curr_weights)):
                    curr_weights[i] = curr_weights[i] * averaging_weight
                layer.set_weights(curr_weights)
            for model in models[1:]:
                averaging_weight = self.score_fn(model(self.golden_data_input),self.golden_data_output)
                sum_weighted_average += averaging_weight
                for (layer0,layer1) in zip(models[0].layers,model.layers):
                    curr_weights = layer0.get_weights()
                    weights = layer1.get_weights()
                    for i in range(len(curr_weights)):
                        curr_weights[i] = curr_weights[i] + averaging_weight*weights[i]
                    layer0.set_weights(curr_weights)
        
            for layer in models[0].layers:
                curr_weights = layer.get_weights()
                for i in range(len(curr_weights)):
                    curr_weights[i] = curr_weights[i] / sum_weighted_average
                layer.set_weights(curr_weights)
            print("After")
            for layer in models[0].layers:
                print(layer.get_weights())
                print("\n")
            return models[0]  
    # def save_model(self,model):
    #     import torch
    #     self.model_name = 'model_' + str(self.server_controller.get_version()) + '.pt'
    #     m = torch.jit.script(model)
    #     torch.jit.save(m, 'torch1.pt')
    #     logging.info("Averaged Model has been saved on Server")
    
    def count_average(self):
        try:
            self.barrier.wait()
        except:
            logging.warning('In count_average, barrier broke')
            self.server_controller.version_updating.acquire()
            pass
        averaged_model = self.calculate_average()
        self.modelsDB.add_model_accuracy(self.score_fn(init_weights(self.golden_data_input),self.golden_data_output), datetime.datetime.now(), len(self.received_values))
        self.server_controller.increase_version()
        whole_path = self.server_model_path + '/' + 'model_' + str(self.server_controller.get_version())
        if self.model_type == "pytorch":
            whole_path += '.pt'
        elif self.model_type == "tensorflow":
            whole_path += '.h5'
        print(whole_path)
        for f in os.listdir(self.server_model_path):
            os.remove(os.path.join(self.server_model_path, f))
        save_model(self.model_type,whole_path,averaged_model)
        # Interrupt server.accept() with fake connection
        fake_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        IP_address = self.ip
        Port = self.port
        fake_server.connect((IP_address, Port))
        for client_socket in self.received_values:
            client_socket.send(b'calculation completed')
        logging.debug("Poison packet has been sent to the DataChannel server")
        self.server_controller.version_updating.release()

    def check_time(self):
        while True:
            if self.last_time is None:
                continue
            current_time = datetime.datetime.now()
            dif = current_time - self.last_time
            dif = dif.total_seconds()
            if dif > self.gap_time:
                self.barrier.abort()
                logging.debug("Barrier Aborted")
                break

    def client_handler(self, client_socket):
        client_socket.send(b'start')
        client = client_socket.recv(1024)
        client = client.decode('utf-8')
        client_socket.send(b'Id Verified')
        logging.info("ID has been received in datachannel server by {}".format(client_socket.getpeername()))
        if self.barrier.broken:
            client_socket.send(b'I won\'t receive connections anymore!')
            return
        # message = int(client_socket.recv(1024).decode())
        model = receive(client_socket,self.model_type)
        logging.info("Model has been received in datachannel server by {}".format(client_socket.getpeername()))
        if client not in self.clientsDB.get_clients():
            client_socket.send(b'I don\'t know you!')
            print("I don't know you!")
            return
        self.last_time = datetime.datetime.now()
        self.received_values[client_socket] = model

    def runner_thread(self):
        while True:
            client_socket, (ip, port) = self.server.accept()
            if self.barrier.broken:
                break
            logging.info('Client connected on datachannel server from {}:{}'.format(ip, port))
            t = threading.Thread(target=self.client_handler, args=(client_socket,))
            t.start()
        self.server.close()



from pymongo import MongoClient

class ClientsRepository:
    def __init__(self,mongodb_host):
        self.con = MongoClient(mongodb_host)
        self.db = self.con.clientsDB
        self.clients = self.db.ids
        # self.clients = []

    def add_client(self, id):
        self.clients.insert_one({'id': id})
        # self.clients.append(id)

    def get_clients(self):
        cur_clients = [client['id'] for client in self.clients.find()]
        return cur_clients
        # return self.clients



import os 
class Validator:
    def __init__(self, yaml_dict):
        self.yaml_dict = yaml_dict
        self.keys = set(['ip', 'datachannel_port', 'ping_port','mongodb_host', 'model_type', 'server_model_path', 'datachannel_time_interval', 'datachannel_gap_time'])
        self.validate()
    
    def validate(self):
        for key in self.keys:
            if key not in self.yaml_dict:
                raise Exception("Key '{}' not found in yaml file".format(key))
        if self.yaml_dict['model_type'] not in ['sklearn', 'tensorflow','pytorch']:
            raise Exception("Model type '{}' not supported".format(self.yaml_dict['model_type']))
        if not os.path.exists(self.yaml_dict['server_model_path']):
            raise Exception("Model path '{}' not found".format(self.yaml_dict['server_model_path']))

        
        



from pymongo import MongoClient

class ModelAccuraciesRepository:
    def __init__(self,mongodb_host):
        self.con = MongoClient(mongodb_host)
        self.db = self.con.clientsDB
        self.models = self.db.models

    def add_model_accuracy(self, accuracy, time, clients):
        self.clients.insert_one({'accuracy': accuracy, 'time': time, 'clients': clients})

    def get_clients_accuracies(self):
        accuracies = [(model['accuracy'], model['time'], model['clients']) for model in self.models.find()]
        return accuracies



from io import BytesIO
import logging
import pickle
from os.path import exists
logging.basicConfig(level=logging.NOTSET)
def prepare_model(model):
    buffer = BytesIO()
    model_type = str(type(model))
    model_parent_type = str(type(model).__bases__)
    if 'keras' in model_type or 'keras' in model_parent_type:
        import tensorflow as tf
        tf.autograph.set_verbosity(1)
        tf.get_logger().setLevel('INFO')
        from tensorflow.keras.models import save_model
        import h5py
        with h5py.File(buffer, 'w') as f:
            save_model(model, f, include_optimizer=True)
    elif 'sklearn' in model_type or 'sklearn' in model_parent_type:
        from joblib import load, dump
        dump(model, buffer)
    elif 'torch' in model_type or 'torch' in model_parent_type:
        import torch
        scripted_model = torch.jit.script(model)
        torch.jit.save(scripted_model, buffer)
    buffer.seek(0)
    logging.debug("Model Prepared")
    return buffer.read()


def load_model(model_type,model_path):
    model = None
    if not exists(model_path):
        return None
    if model_type == "sklearn":
        from joblib import load
        model = load(model_path)
    elif model_type == "tensorflow":
        import tensorflow as tf
        tf.autograph.set_verbosity(1)
        tf.get_logger().setLevel('INFO')
        from tensorflow.keras.models import load_model
        model = load_model(model_path)
    elif model_type == "pytorch":
        import torch
        model = torch.jit.load(model_path)
    logging.info(f"{model_type} Model loaded")  
    return model
    
def save_model(model_type,model_name,model):
    if model_type == 'pytorch':
        import torch
        m = torch.jit.script(model)
        torch.jit.save(m, model_name)
        logging.info("Averaged Model has been saved on Server")
    elif model_type == 'tensorflow':
        import tensorflow as tf
        tf.autograph.set_verbosity(1)
        tf.get_logger().setLevel('INFO')
        from tensorflow.keras.models import save_model
        import h5py
        with h5py.File(model_name, 'w') as f:
            save_model(model, f)
        logging.info("Averaged Model has been saved on Server")

def receive(client_socket,model_type, socket_buffer_size=1024):
    buffer = BytesIO()
    while True:
        data = client_socket.recv(socket_buffer_size)
        if not data:
            break
        buffer.write(data)
        buffer.seek(-4, 2)
        if b'EOF' in buffer.read():
            break
    buffer.seek(0)
    model = load_data(buffer,model_type)
    logging.info("Model Receiving finished")
    return model

def load_data(file: BytesIO,model_type):
    data = file.read()[:-3]
    file.seek(0)
    if model_type == 'sklearn':
        from joblib import load
        return load(file)
    elif model_type == 'tensorflow':
        import tensorflow as tf
        tf.autograph.set_verbosity(1)
        tf.get_logger().setLevel('INFO')
        from tensorflow.keras.models import load_model
        import h5py
        with h5py.File(file, 'r') as f:
            model = load_model(f)
        return model
    elif model_type == 'pytorch':
        import torch
        model = torch.jit.load(file)
        return model

def load_input(input_path,output_path):
    input_data = pickle.load(open(input_path, 'rb'))
    output_data = pickle.load(open(output_path, 'rb'))
    return input_data,output_data


import threading

class ServerController:
    def __init__(self,datachanel_server,ping_server,version=0):
        self.datachanel_server = datachanel_server
        self.ping_server = ping_server
        self.version = version
        self.version_updating = threading.Lock()
        self.version_lock = threading.Lock()
    
    def start(self):
        self.ping_server.start(self)

    def fire(self):
        self.datachanel_server.start(self)

    def get_version(self):
        self.version_lock.acquire()
        cur_version = self.version
        self.version_lock.release()
        return cur_version

    def increase_version(self):
        self.version_lock.acquire()
        self.version = self.version + 1
        self.version_lock.release()
   


from ClientsRepository import *
repo = ClientsRepository("mongodb+srv://doadmin:g12k79jU8L3y0u5t@db-mongodb-nyc3-12601-daeda50b.mongo.ondigitalocean.com/admin?authSource=admin&replicaSet=db-mongodb-nyc3-12601&tls=true")
repo.add_client("f4f35d119d7cd5df3f5a327c92ccc107")


import yaml
import os
from Server.datachannel_server import DatachannelServer
from Server.ping_server import PingServer
from Server.yaml_validator import Validator
from Server.server_controller import ServerController
from Server.ClientsRepository import ClientsRepository
from Server.ModelAccuraciesRepository import ModelAccuraciesRepository
import logging
logging.basicConfig(level=logging.NOTSET)

#178.62.92.57
class AppServer:
    def __init__(self,configuration_path,score_fn) -> None:
        print(configuration_path)
        self.configuration_path = configuration_path
        self.configuration = self.load_configuration()
        self.score_fn = score_fn
        with open("droebit.txt", "w+") as f:
            f.write(str(self.configuration['datachannel_time_interval']))
        print(self.configuration)
    
    def load_configuration(self) -> dict:
        with open(self.configuration_path) as f:
            return yaml.safe_load(f)

    def run(self):
        validator = Validator(self.configuration)
        try:
            validator.validate()
        except Exception as e:
            logging.error(e)
            return
        logging.debug("Configuration is valid")
        clientDB = ClientsRepository(self.configuration['mongodb_host'])
        modelDB = ModelAccuraciesRepository(self.configuration['mongodb_host'])
        ping_server = PingServer(self.configuration['ip'],self.configuration['ping_port'],clientDB,self.configuration['server_model_path'],self.configuration['model_type'], self.configuration['datachannel_time_interval'])
        datachannel_server = DatachannelServer(self.configuration['ip'],self.configuration['datachannel_port'],clientDB,modelDB,self.configuration['server_model_path'],self.configuration['model_type'],self.configuration['golden_data_input_path'],self.configuration['golden_data_output_path'],self.score_fn, self.configuration['datachannel_gap_time'])
        
        path = self.configuration['server_model_path']
        highest_version = 0
        for f in os.listdir(path):
            cur_version = int(f[f.find('_')+1:f.find('.')])
            highest_version = max(highest_version, cur_version)

        server_controller = ServerController(datachannel_server,ping_server,highest_version)
        server_controller.start()

        logging.debug("Server Controller started")


import socket
import threading
import secrets
import datetime
import logging
import time
from Server.utils import prepare_model,load_model
logging.basicConfig(level=logging.NOTSET)


class PingServer:
    def __init__(self, ip, port,clientsDB,server_model_path,model_type,time_interval,starting_time = datetime.datetime.now(), listener_num = 100):
        self.ip = ip
        self.port = port
        self.clientsDB = clientsDB
        self.server_model_path = server_model_path
        self.model_type = model_type
        self.starting_time = starting_time
        self.time_interval = time_interval
        self.listener_num = listener_num
        logging.debug("Ping Server initialized")
        
        
    def start(self,controller):
        self.controller = controller
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind((self.ip, self.port))
        self.server.listen(self.listener_num)
        logging.debug("Ping Server Has been started on {}:{}".format(self.ip, self.port))
        t = threading.Thread(target=self.run)
        t.start()
        t1 = threading.Thread(target=self.time_checker)
        self.is_time = False
        self.is_legal_to_send = True
        t1.start()

    def run(self):
        while True:
            client_socket, (ip, port) = self.server.accept()
            logging.info("Client {} connected on ping server from {}".format(client_socket, ip))
            t = threading.Thread(target=self.client_handler, args=(client_socket,))
            t.start()

    def time_checker(self):
        while True:
            current_time = datetime.datetime.now()
            time_diff = current_time - self.starting_time
            if time_diff.seconds > self.time_interval:
                logging.debug("Datachanel Server Start Fire Called")
                self.is_legal_to_send = False
                self.controller.fire()
                self.starting_time = current_time
                self.is_time = True
            else:
                self.is_time = False
            time.sleep(1)

    def client_handler(self, client_socket, socket_buffer_size=1024):
        initial_message = client_socket.recv(socket_buffer_size).decode()
        logging.debug("Message has been received to Ping Server")
        
        clients = self.clientsDB.get_clients()
        if initial_message == 'Give me an id you son of a bitch!':
            client_id = secrets.token_hex(16)
            while client_id in clients:
                client_id = secrets.token_hex(16)
            self.clientsDB.add_client(client_id)
            client_socket.send(client_id.encode())
            logging.debug("ID has been sent to new Client")
        elif initial_message == '':
            return
        else:
            client_id = initial_message[initial_message.index(':') + 1:]
            if client_id in clients:
                client_socket.send(b'Oh I know you!')
            else:
               client_socket.send(b'Oh I don\'t know you!')
               client_socket.close()
               return


        while True:
            data = client_socket.recv(socket_buffer_size).decode()
            current_time = datetime.datetime.now()
            logging.info('Ping from {} at {}, message: {}'.format(client_id, current_time, data))
            if data == "":
                logging.debug("Client {} disconnected".format(client_id))
                client_socket.close()
                break
            client_version = int(data)
            logging.info('Version of the client\'s model is {}'.format(client_version))
            if client_version < self.controller.get_version():
                self.controller.version_updating.acquire()
                client_socket.send(b'update')
                print(client_socket.recv(1024).decode())
                client_socket.send(str(self.controller.get_version()).encode())
                print(client_socket.recv(1024).decode())
                model_path = self.server_model_path + 'model_' + str(self.controller.get_version()) + '.pt'
                model = load_model(self.model_type,model_path)
                data = prepare_model(model)
                client_socket.sendall(data)
                client_socket.sendall(b'EOF')
                logging.info("Averaged Model Sent to {}".format(client_socket.getpeername()))
                print(client_socket.recv(1024).decode())
                self.controller.version_updating.release()
            elif self.is_time:
                client_socket.send(b'start')
                logging.debug("Sent start signal to the client: {}".format(client_socket.getpeername()))
            else:
                client_socket.send(b'not start')
            


import setuptools

setuptools.setup(
    name="Federated-Learning-Server",
    version="0.0.8",
    author="Bichebi",
    author_email="mgach18@freeuni.edu.ge",
    description="Server package for Collaborative learning framework for deep neural network",
    packages=["Server"],
    url="https://github.com/sHOTa-23/Collaborative-Learning-Framework-for-Deep-Neural-Network"
)
